# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from inspect import signature
from typing import Any

import pytest

import upstage_des.api as UP
from upstage_des.actor import Actor
from upstage_des.base import EnvironmentContext, SimulationError
from upstage_des.states import State


def test_actor_creation() -> None:
    """Ensure instantation of Actor."""
    with EnvironmentContext():
        name = "testing"
        actor = Actor(name=name)
        assert actor.__class__ == Actor
        assert hasattr(actor, "_state_defs")

    with EnvironmentContext():
        a = Actor(name="Three")
        actors = a.get_actors()
        assert len(actors) == 1
        assert a in actors


def test_actor_subclass(
    base_actors: tuple[tuple[UP.State, ...], tuple[type[UP.Actor], ...]],
) -> None:
    """Test subclasses of actor.

    A subclass of actor must have a proper signature and states in its
    dictionary.

    The names of the states also need to match up.

    Any functions must also be a part of the subclass.

    """
    states, actors = base_actors
    first_state, second_state, third_state, fourth_state = states
    ActorSubclass, DoubleSubclass = actors

    # is the new actor a subclass?
    assert issubclass(ActorSubclass, Actor)

    # does the init signature include the values we want?
    actor_signature = signature(ActorSubclass.__init__)
    expected_params = ["state_one", "state_two", "name"]
    for parameter in expected_params:
        assert parameter in actor_signature.parameters
    assert "kwargs" not in actor_signature.parameters

    # test making an instance without arguments raising an error
    with pytest.raises(Exception):
        ActorSubclass()  # type: ignore [call-arg]

    with EnvironmentContext():
        # create an instance
        instance = ActorSubclass(
            name="Testing",
            state_one=1,
            state_two=2,
        )

        assert instance.a_function(123) == (instance, 123)  # type: ignore [attr-defined]

        # test the state definitions
        assert hasattr(instance, "_state_defs")
        assert len(instance._state_defs) == 2
        assert instance._state_defs["state_one"] is first_state
        assert instance._state_defs["state_two"] is second_state
        assert repr(instance) == "ActorSubclass: Testing"
        sts = instance.state_values
        exp = {"state_one": 1, "state_two": 2}
        assert sts == exp

    # Test that copying a state name will cause a failure.
    with pytest.raises(ValueError):

        class _(DoubleSubclass):  # type: ignore [valid-type, misc]
            state_three = UP.State[float](default=1.2)


def test_multiple_inheritance(
    base_actors: tuple[tuple[UP.State, ...], tuple[type[UP.Actor], ...]],
) -> None:
    """Test actor subclasses but for an additional subclass."""
    states, actors = base_actors
    first_state, second_state, third_state, fourth_state = states
    ActorSubclass, DoubleSubclass = actors

    assert issubclass(DoubleSubclass, Actor)
    assert issubclass(DoubleSubclass, ActorSubclass)

    actor_signature = signature(DoubleSubclass.__init__)
    expected_params = ["state_one", "state_two", "state_three", "name"]
    for parameter in expected_params:
        assert parameter in actor_signature.parameters
    assert "kwargs" not in actor_signature.parameters

    with EnvironmentContext():
        instance = DoubleSubclass(
            name="Testing",
            state_one=1,
            state_two=2,
            state_three=3,
            state_four=4,
        )
        assert instance.b_function(123) == (instance, 123)  # type: ignore [attr-defined]

        # test the state definitions
        assert hasattr(instance, "_state_defs")
        assert len(instance._state_defs) == 4
        assert "state_three" in instance.states
        assert "state_one" in instance.states
        assert instance._state_defs["state_one"] is first_state
        assert instance._state_defs["state_three"] is third_state
        assert instance._state_defs["state_four"] is fourth_state


def test_get_knowledge() -> None:
    class TestActor(Actor):
        pass

    with EnvironmentContext():
        act = TestActor(name="A Test Actor")

        name = "new"
        other_name = "some data"
        value = {"A": 1, "B": 2}
        act.set_knowledge(name, value)

        returned_value = act.get_knowledge(name)
        assert value == returned_value, "Returned value is not the same knowledge"

        other_value = act.get_knowledge(other_name)
        assert other_value is None, "Empty knowledge returned something other than None"


def test_set_knowledge() -> None:
    class TestActor(Actor):
        pass

    with EnvironmentContext():
        act = TestActor(name="A Test Actor")

        name = "new"
        value = {"A": 1, "B": 2}
        value2 = {"A": 5, "B": 6}
        act.set_knowledge(name, value)

        with pytest.raises(SimulationError):
            act.set_knowledge(name, value2)

        act.set_knowledge(name, value2, overwrite=True)
        returned_value = act.get_knowledge(name)
        assert value2 == returned_value, "Returned value is not the same knowldge"


def test_clear_knowledge() -> None:
    class TestActor(Actor):
        pass

    with EnvironmentContext():
        act = TestActor(name="A Test Actor")

        name = "new"
        value = {"A": 1, "B": 2}
        act.set_knowledge(name, value)

        act.get_knowledge(name)

        act.clear_knowledge(name)
        know = act.get_knowledge(name)
        assert know is None, "Knowledge was not cleared"


def test_get_and_clear() -> None:
    class TestActor(Actor):
        pass

    with EnvironmentContext():
        act = TestActor(name="Second test")
        act.set_knowledge("thing", {3: 1})
        v = act.get_and_clear_knowledge("thing")
        assert v == {3: 1}
        assert "thing" not in act._knowledge

        with pytest.raises(UP.SimulationError):
            act.get_and_clear_knowledge("thing")

        t = UP.Task()
        act.set_knowledge("other", {2: 3})
        v = t.get_and_clear_actor_knowledge(act, "other")
        assert v == {2: 3}
        assert "other" not in act._knowledge


def test_bulk_knowledge() -> None:
    class TestActor(Actor):
        pass

    with UP.EnvironmentContext():
        know = {"one": 1, "two": 2}
        act = TestActor(name="Example")
        act.set_bulk_knowledge(know)
        assert know == act._knowledge

        with pytest.raises(UP.SimulationError):
            act.set_bulk_knowledge({"one": 3, "three": 3})

        act.set_bulk_knowledge({"one": 11, "three": 3}, overwrite=True)
        v = act.get_bulk_knowledge(set(["one", "two", "three"]))
        assert v == {"one": 11, "two": 2, "three": 3}

        v = act.get_and_clear_bulk_knowledge(["two", "three"])
        assert v == {"two": 2, "three": 3}
        assert act._knowledge == {"one": 11}

        t = UP.Task()
        t.set_actor_bulk_knowledge(act, know, overwrite=True)
        assert act._knowledge == know
        v = t.get_actor_bulk_knowledge(act, ["one", "two"])
        assert v == know

        with pytest.raises(UP.SimulationError):
            t.set_actor_bulk_knowledge(act, {"one": 3, "three": 3})

        v = t.get_and_clear_actor_bulk_knowledge(act, ["one", "two"])
        assert v == know
        assert act._knowledge == {}


def test_knowledge_event() -> None:
    with EnvironmentContext() as env:
        act = Actor(name="A test actor")
        evt = act.create_knowledge_event(name="Waiter")
        assert act._knowledge.get("Waiter", None) is evt
        assert evt.is_complete() is False
        act.succeed_knowledge_event(name="Waiter")
        env.run()
        assert act._knowledge.get("Waiter", None) is None
        assert evt.is_complete()


def test_actor_copying() -> None:
    class SomeActor(Actor):
        kind = "a simple actor for testing"
        some_state = State[Any]()

    with EnvironmentContext():
        actor = SomeActor(name="some actor", some_state=True)

        clone = actor.clone(new_env=None, some_state=False)

        assert clone.some_state != actor.some_state

        assert actor.name in clone.name

        assert "[CLONE" in clone.name

        assert actor._num_clones == 1

        assert actor.kind == clone.kind


def test_actor_copy_with_knowledge() -> None:
    class SomeActor(Actor):
        kind = "a simple actor for testing"
        some_state = State[Any]()

    with EnvironmentContext():
        actor = SomeActor(name="some actor", some_state=True)
        d_values = {"A": 1, "B": 2}
        float_value = 1234.567

        actor.set_knowledge("new", d_values)
        actor.set_knowledge("other", float_value)

        clone = actor.clone(new_env=None, some_state=False)
        for name in ["new", "other"]:
            v1 = actor.get_knowledge(name)
            v2 = clone.get_knowledge(name)
            assert v1 == v2, "Copied knowledge is different"

        # we can't test for equal IDs or object equivalence because of how
        # Python handles memory for booleans, small integers, etc.

        d_values["A"] = 23
        v1 = actor.get_knowledge("new")
        assert v1["A"] == 23, "Input knowledge did not retain reference"

        v2 = clone.get_knowledge("new")
        assert v2["A"] == 1, "Cloned knowledge retained reference"



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from typing import Any, cast

import pytest
from simpy import Container, Environment, Store

import upstage_des.api as UP
import upstage_des.resources.monitoring as monitor
from upstage_des.actor import Actor
from upstage_des.api import EnvironmentContext, SimulationError, UpstageError
from upstage_des.states import LinearChangingState, ResourceState, State
from upstage_des.type_help import SIMPY_GEN


class StateTest:
    state_one = State[Any]()
    state_two = State[Any](recording=True)

    def __init__(self, env: Environment | None) -> None:
        cast(UP.Actor, self)
        self.env = env
        # including for compatibility
        self._mimic_states: dict[str, Any] = {}
        self._state_listener = None
        self._state_histories: dict[str, list[tuple[float, Any]]] = {}

    def set_one(self, val: Any) -> None:
        self.state_one = val  # type: ignore [arg-type]

    def set_two(self, val: Any) -> None:
        self.state_two = val  # type: ignore [arg-type]


class StateTestActor(Actor):
    state_one = State[Any]()
    state_two = State[Any](recording=True)
    state_three = LinearChangingState(recording=True)


class MutableDefaultActor(Actor):
    lister = State[list](default=[])
    diction = State[dict](default={})
    setstate = State[set](default=set())


def test_state_fails_without_env() -> None:
    """Test that recording states need the class to have an env attribute"""
    tester = StateTest(None)
    # the first one should not raise an error
    tester.set_one(1)

    with pytest.raises(SimulationError):
        tester.set_two(1)


def test_state_values() -> None:
    """Test that we get the right state values we input"""
    with EnvironmentContext(initial_time=1.5) as env:
        tester = StateTest(env)
        tester.state_one = 1  # type: ignore [arg-type]
        assert tester.state_one == 1  # type: ignore [arg-type]
        tester.state_two = 2  # type: ignore [arg-type]
        assert tester.state_two == 2  # type: ignore [arg-type]


def test_state_recording() -> None:
    with EnvironmentContext(initial_time=1.5) as env:
        tester = StateTest(env)
        tester.state_two = 2  # type: ignore [arg-type]
        assert "state_two" in tester._state_histories
        env.run(until=2.5)
        tester.state_two = 3  # type: ignore [arg-type]
        assert len(tester._state_histories["state_two"]) == 2
        assert tester._state_histories["state_two"][0] == (1.5, 2)
        assert tester._state_histories["state_two"][1] == (2.5, 3)


def test_state_mutable_default() -> None:
    with EnvironmentContext(initial_time=1.5):
        tester = MutableDefaultActor(name="Example")
        tester2 = MutableDefaultActor(name="Example2")
        assert id(tester.lister) != id(tester2.lister)
        tester.lister.append(1)
        assert len(tester2.lister) == 0
        assert len(tester.lister) == 1

        assert id(tester.diction) != id(tester2.diction)
        tester2.diction[1] = 2
        assert len(tester.diction) == 0
        assert len(tester2.diction) == 1

        assert id(tester.setstate) != id(tester2.setstate)
        tester2.setstate.add(1)
        assert len(tester.setstate) == 0
        assert len(tester2.setstate) == 1


def test_state_values_from_init() -> None:
    with EnvironmentContext() as env:
        tester = StateTestActor(
            name="testing",
            state_one=1,
            state_two=2,
            state_three=4,
        )
        env.run(until=1.5)
        assert tester.state_one == 1
        assert tester.state_two == 2
        assert tester.state_three == 4
        tester.state_three = 3
        assert "state_three" in tester._state_histories
        assert tester._state_histories["state_three"] == [(0.0, 4), (1.5, 3)]


def test_state_none_allowed() -> None:
    class NoneActor(Actor):
        example = State[int](allow_none_default=True)

    with EnvironmentContext():
        ex = NoneActor(name="Example")

        with pytest.raises(SimulationError, match="State example should have been set"):
            ex.example

        ex.example = 3
        assert ex.example == 3


def test_linear_changing_state() -> None:
    state_three_init = 3
    init_time = 1.5
    rate = 3.1
    timestep = 1

    with EnvironmentContext(initial_time=init_time) as env:
        tester = StateTestActor(
            name="testing",
            state_one=1,
            state_two=2,
            state_three=state_three_init,
        )

        task = UP.Task()

        tester.activate_state(state="state_three", task=task, rate=rate)
        assert "state_three" in tester._active_states
        assert tester._active_states["state_three"] == tester.get_active_state_data(
            "state_three", without_update=True
        )
        env.run(until=init_time + timestep)
        # Test getting the value before it's ended
        assert tester.state_three == rate * timestep + state_three_init
        env.run(until=init_time + timestep * 2)
        state_data = tester.get_active_state_data("state_three", without_update=True)
        assert state_data["started_at"] == timestep + init_time
        assert state_data["rate"] == rate
        tester.deactivate_state(state="state_three", task=task)
        assert "state_three" not in tester._active_states
        assert tester.state_three == rate * timestep * 2 + state_three_init


def test_resource_state_valid_types() -> None:
    class Holder(Actor):
        res = ResourceState[Store](valid_types=Store)

    with EnvironmentContext():
        Holder(
            name="example",
            res={"kind": Store},
        )

        with pytest.raises(UpstageError):
            Holder(
                name="example",
                res={"kind": Container},
            )

        with pytest.raises(UpstageError):

            class _(Actor):
                res = ResourceState[Store](valid_types=(1,))  # type: ignore [arg-type]

        with pytest.raises(UpstageError):

            class _(Actor):  # type: ignore [no-redef]
                res = ResourceState[Store](valid_types=(Actor,))


def test_resource_state_set_protection() -> None:
    class Holder(Actor):
        res = ResourceState[Store](valid_types=(Store))

    with EnvironmentContext():
        h = Holder(
            name="example",
            res={"kind": Store},
        )
        with pytest.raises(UpstageError, match=".+It cannot be changed once set.+"):
            h.res = 1.0


def test_resource_state_no_default_init() -> None:
    class Holder(Actor):
        res = ResourceState[Store]()

    with EnvironmentContext():
        with pytest.raises(UpstageError, match="Missing values for states"):
            h = Holder(
                name="example",
            )

        with pytest.raises(UpstageError, match="No resource type"):
            h = Holder(
                name="example",
                res={},
            )

        h = Holder(
            name="example",
            res={"kind": Store},
        )
        assert isinstance(h.res, Store)


def test_resource_state_default_init() -> None:
    class Holder(Actor):
        res = ResourceState[Store](default=Store)
        res2 = ResourceState[Container](
            default=Container, default_kwargs={"capacity": 11, "init": 5}
        )

    class HolderBad(Actor):
        res = ResourceState[Store](default=Store)
        res2 = ResourceState[Container](default=Container, default_kwargs={"capa": 11, "init": 5})

    with EnvironmentContext():
        h = Holder(name="Example")
        assert isinstance(h.res, Store)
        assert h.res2.capacity == 11
        assert h.res2.level == 5

        h = Holder(name="Example", res={"capacity": 10}, res2={"capacity": 12})
        assert isinstance(h.res, Store)
        assert h.res.capacity == 10
        assert h.res2.capacity == 12
        assert h.res2.level == 5

        with pytest.raises(UpstageError):
            HolderBad(name="Bad one")


def test_resource_state_kind_init() -> None:
    class Holder(Actor):
        res = ResourceState[Store]()

    with EnvironmentContext():
        h = Holder(name="Example", res={"kind": Store, "capacity": 10})
        assert isinstance(h.res, Store)
        assert h.res.capacity == 10

        h = Holder(name="Example", res={"kind": Container, "capacity": 100, "init": 50})
        assert isinstance(h.res, Container)
        assert h.res.capacity == 100
        assert h.res.level == 50

        test_resources = [
            x
            for x in monitor.__dict__.values()
            if isinstance(x, type) and issubclass(x, Store | Container)
        ]
        for the_class in test_resources:
            h = Holder(name="Example", res={"kind": the_class, "capacity": 99})
            assert isinstance(h.res, the_class)
            assert h.res.capacity == 99


def test_resource_state_simpy_store_running() -> None:
    class Holder(Actor):
        res = ResourceState[Store]()

    with EnvironmentContext() as env:
        h = Holder(name="Example", res={"kind": Store, "capacity": 10})

        def put_process(entity: Holder) -> SIMPY_GEN:
            for i in range(11):
                yield env.timeout(1.0)
                yield entity.res.put(f"Item {i}")
            return "Done"

        def get_process(entity: Holder) -> SIMPY_GEN:
            res = yield entity.res.get()
            return res

        proc_1 = env.process(put_process(h))
        proc_2 = env.process(get_process(h))
        env.run()
        assert proc_2.value == "Item 0"
        assert proc_1.value == "Done"


def test_resource_clone() -> None:
    class Holder(Actor):
        res = ResourceState[Store](default=Store)

    class Holder2(Actor):
        res = ResourceState[Container](default=Container)

    with EnvironmentContext():
        holder = Holder(name="example")
        holder_2 = holder.clone()
        assert id(holder_2.res.items) != id(holder.res.items)

        holder = Holder2(name="example")
        holder_2 = holder.clone()
        assert id(holder_2.res.level) != id(holder.res.level)


class HelperCallback:
    def __init__(self) -> None:
        self.cbacks: list[tuple[Any, Any]] = []

    def _callbacker(self, instance: Any, value: Any) -> None:
        self.cbacks.append((instance, value))


def test_state_callback() -> None:
    class CbackActor(Actor):
        state_one = State[Any](recording=True)

    helper = HelperCallback()
    with EnvironmentContext():
        actor = CbackActor(
            name="Test",
            state_one=1,
        )

        actor._add_callback_to_state("source", helper._callbacker, "state_one")
        actor.state_one = 2
        assert len(helper.cbacks) == 1
        assert helper.cbacks[0][1] == 2
        actor._remove_callback_from_state("source", "state_one")

        actor.state_one = 3
        assert len(helper.cbacks) == 1


def test_matching_states() -> None:
    """Test the state matching code.
    At this time, state matching only works with CommunicationStore. It's the
    only state with a special attribute attached to it.
    """

    class Worker(UP.Actor):
        sleepiness = UP.State[float](default=0.0, valid_types=(float,))
        walkie = UP.CommunicationStore(mode="UHF")
        intercom = UP.CommunicationStore(mode="loudspeaker")

    with EnvironmentContext():
        worker = Worker(name="Billy")
        store_name = worker._get_matching_state(
            UP.CommunicationStore,
            {"_mode": "loudspeaker"},
        )
        assert store_name is not None
        store = getattr(worker, store_name, "")
        assert store is worker.intercom, "Wrong state retrieved"
        assert store is not worker.walkie, "Wrong state retrieved"

        # Show the FCFS behavior
        state_name = worker._get_matching_state(
            UP.State,
        )
        assert state_name is not None
        value = getattr(worker, state_name)
        assert value == worker.sleepiness, "Wrong state retrieved"

        # Show the FCFS behavior with state type
        state_name = worker._get_matching_state(
            UP.CommunicationStore,
        )
        assert state_name is not None
        value = getattr(worker, state_name)
        assert value is worker.walkie, "Wrong state retrieved"



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import simpy as SIM
from simpy import Environment, Process

from upstage_des.actor import Actor
from upstage_des.base import EnvironmentContext, MockEnvironment
from upstage_des.constants import PLANNING_FACTOR_OBJECT
from upstage_des.events import Any, Get, Put, ResourceHold, Wait
from upstage_des.states import LinearChangingState, State
from upstage_des.task import Task
from upstage_des.type_help import SIMPY_GEN, TASK_GEN


class ActorForTest(Actor):
    dummy = State[list]()


class StateActor(Actor):
    fuel = LinearChangingState(recording=True)
    fuel_burn = State[float](recording=True)


class Flight:
    def __init__(self, code: int) -> None:
        self.code = code
        self.dummy: list = []


class AirplaneTask(Task):
    limit: float
    rate: float
    store: SIM.Store
    orders_time: float
    reason: list

    def task(self, *, actor: StateActor) -> TASK_GEN:
        # this task could be thought of as loitering, waiting for a get
        # request, interrupt, or prescribed failure
        time_to_leave = (actor.fuel - self.limit) / self.rate

        leave_event = Wait(time_to_leave)
        orders_event = Get(self.store, rehearsal_time_to_complete=self.orders_time)

        actor.activate_linear_state(state="fuel", task=self, rate=-self.rate)
        any_event = Any(leave_event, orders_event)
        yield any_event
        actor.deactivate_state(state="fuel", task=self)

        if orders_event.is_complete():
            self.reason.append(orders_event.get_value())
        else:
            orders_event.cancel()
        if leave_event.is_complete():
            self.reason.append("LEAVE")


class BigTask(Task):
    time: float
    maintenance_bay: SIM.Resource
    broken_vehicle_depot: SIM.Store
    fixed_vehicle_depot: SIM.Store

    def task(
        self,
        *,
        actor: Flight,
    ) -> TASK_GEN:
        # a function to mimic extra code needed to perform a task
        # TODO: This should be wrapped to handle the planning answer
        def test_flight(flight: Flight, planning_answer: float = 3.0) -> float:
            if flight is PLANNING_FACTOR_OBJECT:
                return planning_answer
            return flight.code * 1.5

        yield Wait(self.time)
        actor.dummy.append(self.time)

        # wait for a resource, then do stuff, then give it back
        # get a bay for the actor to do work in
        resource_event = ResourceHold(self.maintenance_bay)
        yield resource_event

        # get a vehicle to fix
        vehicle_to_fix = yield Get(self.broken_vehicle_depot)

        # fix it
        time_to_fix = test_flight(vehicle_to_fix, planning_answer=6.0)
        yield Wait(time_to_fix)
        actor.dummy.append(time_to_fix)

        # put it in fixed
        yield Put(self.fixed_vehicle_depot, vehicle_to_fix)

        # leave the job site
        yield resource_event


def interrupting_task(*, env: Environment, time: float, other_task: Process) -> SIMPY_GEN:
    yield env.timeout(time)
    if other_task.is_alive:
        other_task.interrupt("cancelling")


def test_event_store_returns() -> None:
    with EnvironmentContext() as env:
        store = SIM.Store(env, capacity=1)
        put_object = ("A Test Object", 1.0)
        store.put(put_object)
        env.run()

        class DoARun(Task):
            def task(self, *, actor: ActorForTest) -> TASK_GEN:
                return_value = yield Get(store)
                actor.dummy.append(return_value)

        actor = ActorForTest(name="testing", dummy=[])
        actor_2 = DoARun().rehearse(
            actor=actor,
        )

        assert actor_2.dummy[0] is PLANNING_FACTOR_OBJECT

        _ = DoARun().run(
            actor=actor,
        )
        env.run()
        assert actor.dummy[0] is put_object, "Object returned is not the correct object"


def test_task_with_all_events() -> None:
    with EnvironmentContext() as env:
        store = SIM.Store(env, capacity=2)
        store2 = SIM.Store(env, capacity=2)
        resource = SIM.Resource(env)

        f = Flight(2)
        f2 = Flight(3)
        store.put(f)
        store.put(f2)
        env.run()

        actor = ActorForTest(name="maintenance person", dummy=[])

        # test the task
        bt = BigTask()
        bt.time = 2.0
        bt.broken_vehicle_depot = store
        bt.maintenance_bay = resource
        bt.fixed_vehicle_depot = store2
        test_actor = bt.rehearse(
            actor=actor,
        )

        # check that the returned test actor has the expected entries in its state
        assert test_actor.dummy[0] == 2.0, "Wrong result in test actor"
        assert test_actor.dummy[1] == 6.0, "Wrong result in test actor for fake store object"
        assert f in store.items, "Item was removed when it shouldn't have been"
        assert f2 in store.items, "Item was removed when it shouldn't have been"
        # run the process for real
        bt = BigTask()
        bt.time = 2.0
        bt.broken_vehicle_depot = store
        bt.maintenance_bay = resource
        bt.fixed_vehicle_depot = store2
        _ = bt.run(
            actor=actor,
        )

        env.run()
        assert actor.dummy[0] == 2.0, "Wrong result in test actor"
        assert actor.dummy[1] == f.code * 1.5, "Wrong result in test actor for fake store object"
        assert f not in store.items, "Item wasn't removed when it should have been"
        assert f2 in store.items, "Item was removed when it shouldn't have been"
        assert f in store2.items, "Item wasn't moved to the next store"


def test_task_with_get() -> None:
    with EnvironmentContext() as env:
        orders = SIM.Store(env)
        actor = StateActor(name="Airplane", fuel=100, fuel_burn=5.2)
        result = []
        order_time = 12.3

        class SimpleTask(Task):
            def task(self, *, actor: StateActor) -> TASK_GEN:
                event = Get(orders, rehearsal_time_to_complete=order_time)
                res = yield event
                result.append(res)
                result.append(self.env.now)
                result.append(event)
                result.append(self.env)

        SimpleTask().rehearse(
            actor=actor,
        )

        assert isinstance(result[3], MockEnvironment), f"Not a mock environment: {result[3]}"
        assert result[0] is PLANNING_FACTOR_OBJECT
        assert result[1] == 12.3
        assert result[2].is_complete, "Tested event believes it completed"


def test_task_rehearsal_with_cancels() -> None:
    with EnvironmentContext() as env:
        orders = SIM.Store(env)
        actor = StateActor(name="Airplane", fuel=100, fuel_burn=5.2)

        at = AirplaneTask()
        at.rate = 1.2
        at.limit = 5
        at.store = orders
        at.orders_time = 0.0
        at.reason = []

        tested_actor = at.rehearse(actor=actor)

        assert at.reason[0] == PLANNING_FACTOR_OBJECT
        assert len(at.reason) == 1, "Wrong number of values returned"
        assert tested_actor.fuel == 100

        at = AirplaneTask()
        at.rate = 1.2
        at.limit = 5
        at.store = orders
        at.orders_time = 90.0
        at.reason = []
        tested_actor = at.rehearse(
            actor=actor,
        )

        assert at.reason[0] == "LEAVE"
        assert len(at.reason) == 1, "Wrong number of values returned"
        assert tested_actor.fuel == 5


def test_task_with_cancels() -> None:
    with EnvironmentContext() as env:
        orders = SIM.Store(env)
        actor = StateActor(name="Airplane", fuel=100, fuel_burn=5.2)
        reason: list[str] = []

        at = AirplaneTask()
        at.rate = 1.2
        at.limit = 5
        at.store = orders
        at.orders_time = 0.0
        at.reason = reason

        at.run(
            actor=actor,
        )

        # run the task
        env.run()
        assert reason[0] == "LEAVE", "Wrong leave reason"
        assert env.now == (100 - 5) / 1.2, "Wrong environment time"
        assert actor.fuel == 5, "Wrong fuel"

        # test the task when orders are given
        def give_orders(env: SIM.Environment, time: float, orders: SIM.Store) -> SIMPY_GEN:
            yield env.timeout(time)
            yield orders.put("STOP WHAT YOU ARE DOING")

        ##############
    with EnvironmentContext() as env:
        orders = SIM.Store(env)
        actor = StateActor(name="Airplane", fuel=100, fuel_burn=5.2)

        at = AirplaneTask()
        at.rate = 1.2
        at.limit = 5
        at.store = orders
        at.orders_time = 0.0
        at.reason = []

        at.run(
            actor=actor,
        )
        _ = env.process(give_orders(env, 23.0, orders))
        env.run()
        assert at.reason[0] == "STOP WHAT YOU ARE DOING"

        ##############
        # test the task when orders are given, but it's interrupted beforehand
    with EnvironmentContext() as env:
        orders = SIM.Store(env)
        actor = StateActor(name="Airplane", fuel=100, fuel_burn=5.2)

        at = AirplaneTask()
        at.rate = 1.2
        at.limit = 5
        at.store = orders
        at.orders_time = 0.0
        at.reason = []

        task_proc = at.run(
            actor=actor,
        )

        _ = env.process(give_orders(env, 23.0, orders))
        _ = env.process(interrupting_task(env=env, time=16.5, other_task=task_proc))
        env.run(until=60)

        # no reason should be given
        assert not at.reason
        expected_fuel = 100 - (16.5 * 1.2)
        assert actor.fuel == expected_fuel
        assert len(env._queue) == 1, "End environment queue is too long"
        assert env._queue[0][3].callbacks == [], "Timeout had callbacks that should be cleared"
